import os
import tempfile

import django.dispatch
import pzip
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.files.storage import FileSystemStorage
from django.utils.encoding import force_bytes

__version__ = "1.2.0"
__version_info__ = tuple(int(num) for num in __version__.split("."))


needs_rotation = django.dispatch.Signal()
needs_encryption = django.dispatch.Signal()
bad_keys = django.dispatch.Signal()


class IntermediateFile:
    def __init__(self, path):
        self.path = path

    def temporary_file_path(self):
        return self.path


class PZipStorage(FileSystemStorage):
    DEFAULT_EXTENSION = ".pz"

    DEFAULT_NOCOMPRESS = set(
        [
            ".z",
            ".gz",
            ".zip",
            ".zipx",
            ".tgz",
            ".jpg",
            ".jpeg",
            ".png",
            ".gif",
            ".sit",
            ".sitx",
            ".7z",
            ".pz",
            ".bz2",
            ".xz",
            ".rar",
            ".cab",
            ".zst",
            ".lz",
            ".lzma",
            ".lz4",
            ".jar",
            ".heic",
            ".avif",
            ".webp",
        ]
    )

    def __init__(self, *args, **kwargs):
        self.keys = kwargs.pop(
            "keys", getattr(settings, "PZIP_STORAGE_KEYS", self.default_keys)
        )
        self.extension = kwargs.pop(
            "extension",
            getattr(settings, "PZIP_STORAGE_EXTENSION", self.DEFAULT_EXTENSION),
        )
        self.nocompress = kwargs.pop(
            "nocompress",
            getattr(settings, "PZIP_STORAGE_NOCOMPRESS", self.DEFAULT_NOCOMPRESS),
        )
        if not self.keys:
            raise ImproperlyConfigured("PZipStorage requires at least one key.")
        super().__init__(*args, **kwargs)

    def default_keys(self):
        yield force_bytes(settings.SECRET_KEY)

    def iter_keys(self):
        keys = self.keys() if callable(self.keys) else self.keys
        yield from keys

    def is_pzip(self, name):
        try:
            pzip.open(self.path(name)).close()
            return True
        except pzip.InvalidFile:
            return False

    def size(self, name):
        try:
            with pzip.open(self.path(name)) as f:
                return f.plaintext_size()
        except pzip.InvalidFile:
            return super().size(name)

    def _open(self, name, mode="rb"):
        if self.is_pzip(name):
            assert mode == "rb"  # TODO: support writable files here?
            for idx, key in enumerate(self.iter_keys()):
                try:
                    f = super()._open(name, mode)
                    pz = pzip.open(f, mode, key=key, peek=True)
                    if idx > 0:
                        # If we opened this file with an old key, broadcast a signal for
                        # callers to do rotation.
                        needs_rotation.send(
                            sender=self.__class__, storage=self, name=name, key=key
                        )
                    return pz
                except pzip.InvalidFile:
                    # Close the underlying fileobj if PZip fails to decode.
                    f.close()
            # If we tried all the keys and haven't returned yet for a PZip file, send a
            # bad_keys signal.
            bad_keys.send(sender=self.__class__, storage=self, name=name)
        else:
            # Send needs_encryption signal for non-pzip files.
            needs_encryption.send(sender=self.__class__, storage=self, name=name)
        return super()._open(name, mode)

    def create_encrypted_file(self, name, content):
        try:
            # Use the first (most recent) defined key for encryption.
            key = next(self.iter_keys())
        except StopIteration:
            raise ImproperlyConfigured("PZipStorage requires at least one key.")

        # Determine whether we should compress based on the original supplied file
        # extension.
        should_compress = os.path.splitext(name)[1].lower() not in self.nocompress

        # Create a temporary file to do the encryption/compression before handing off.
        fd, path = tempfile.mkstemp(
            suffix=self.extension, dir=settings.FILE_UPLOAD_TEMP_DIR
        )
        with pzip.open(fd, "wb", key=key, compress=should_compress) as f:
            for chunk in content.chunks():
                f.write(chunk)

        return path

    def _save(self, name, content):
        if not isinstance(content, IntermediateFile):
            path = self.create_encrypted_file(name, content)
            content = IntermediateFile(path)

        # FileSystemStorage will just move the file we wrote, so no need to clean up.
        return super()._save(name + self.extension, content)



import os
import tempfile
import unittest
from unittest.mock import MagicMock

import pzip
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.files.base import ContentFile

from pzip_storage import PZipStorage, bad_keys, needs_encryption, needs_rotation


class CompressedStorageTests(unittest.TestCase):
    def setUp(self):
        self.storage = PZipStorage()

    def test_improper_config(self):
        with self.assertRaises(ImproperlyConfigured):
            PZipStorage(keys=[])
        # PZipStorage does not check callables at creation time.
        storage = PZipStorage(keys=lambda: [])
        with self.assertRaises(ImproperlyConfigured):
            storage.save("testfile", ContentFile(b"test data"))

    def test_round_trip(self):
        plaintext = b"Hello world!"
        name = self.storage.save("hello.txt", ContentFile(plaintext))
        with self.storage.open(name) as f:
            self.assertEqual(plaintext, f.read())
        self.assertEqual(len(plaintext), self.storage.size(name))
        self.assertTrue(self.storage.exists(name))
        self.storage.delete(name)
        self.assertFalse(self.storage.exists(name))

    def test_multiple_keys(self):
        plaintext = (
            b"Answer to the Ultimate Question of Life, The Universe, and Everything."
        )
        keys = [b"first"]
        handler = MagicMock()
        needs_rotation.connect(handler, sender=PZipStorage)
        storage = PZipStorage(keys=lambda: keys)
        name = storage.save("testfile", ContentFile(plaintext))
        keys.insert(0, b"second")
        keys.insert(0, b"third")
        with storage.open(name) as f:
            self.assertEqual(plaintext, f.read())
            handler.assert_called_once_with(
                signal=needs_rotation,
                sender=PZipStorage,
                storage=storage,
                name=name,
                key=b"first",
            )
        storage.delete(name)

    def test_no_compression(self):
        name = self.storage.save("test.jpg", ContentFile(b"JPEG data"))
        with self.storage.open(name) as f:
            self.assertIsInstance(f, pzip.PZip)
            self.assertEqual(f.compression, pzip.Compression.NONE)

    def test_unencrypted(self):
        handler = MagicMock()
        needs_encryption.connect(handler, sender=PZipStorage)
        self.assertEqual(self.storage.size("unencrypted"), 11)
        with self.storage.open("unencrypted") as f:
            self.assertNotIsInstance(f, pzip.PZip)
            self.assertEqual(f.read(), b"hello world")
            handler.assert_called_once_with(
                signal=needs_encryption,
                sender=PZipStorage,
                storage=self.storage,
                name="unencrypted",
            )

    def test_bad_keys(self):
        handler = MagicMock()
        bad_keys.connect(handler, sender=PZipStorage)
        with self.storage.open("encrypted" + PZipStorage.DEFAULT_EXTENSION) as f:
            self.assertNotEqual(f.read(), b"unrecoverable data")
            handler.assert_called_once_with(
                signal=bad_keys,
                sender=PZipStorage,
                storage=self.storage,
                name="encrypted" + PZipStorage.DEFAULT_EXTENSION,
            )


if __name__ == "__main__":
    with tempfile.TemporaryDirectory() as tempdir:
        # Write a pre-existing unencrypted file to the storage root.
        with open(os.path.join(tempdir, "unencrypted"), "wb") as f:
            f.write(b"hello world")
        # Write a pre-existing encrypted file (with a random key) to the storage root.
        random_key = os.urandom(32)
        with pzip.open(
            os.path.join(tempdir, "encrypted" + PZipStorage.DEFAULT_EXTENSION),
            "wb",
            key=random_key,
        ) as f:
            f.write(b"unrecoverable data")
        # Set up Django settings to have a stable SECRET_KEY and MEDIA_ROOT.
        settings.configure(SECRET_KEY=os.urandom(32), MEDIA_ROOT=tempdir)
        unittest.main()
